// not part of package InfoCollection
import InfoCollection.SamplingRule;
import InfoCollection.StoppingRuleSeq;
import InfoCollection.OneStoppingRule;
import InfoCollection.StoppingRule;
import InfoCollection.FixedStop;
import InfoCollection.Simulator;
import InfoCollection.TestExamples;
import InfoCollection.NormalTruth;
import InfoCollection.Truth;
import InfoCollection.FixedStop;
import InfoCollection.IndependentNormalVarianceKnown.*;
import InfoCollection.IndependentNormalVarianceKnown.Factory.*;
import java.lang.reflect.Method;
import java.lang.reflect.InvocationTargetException;
import java.lang.StrictMath;

public class Sim200904 {
  // change the following from 1 to 0 to run more samples for the P3 tests.
  // This will run much longer, but will reproduce the tables in the paper. 
  static boolean do_production = false;

	/*
	 * This is the basic code for running the experiment, in various versions.
	 */

	/* Just run one stopping rule/policy. */
	private static double[] Run(Belief prior, SamplingRule policy, StoppingRule stop, double logc, Boolean doprint, int nruns, int maxN) {
		double[][] ret;
		StoppingRuleSeq stopSeq = new OneStoppingRule(stop);
		double logc_list[] = new double[1];
		logc_list[0] = logc;
		ret = Run(prior,policy,stopSeq,logc_list,doprint,nruns,maxN);
		double[] ret0 = new double[ret[0].length];
		for(int i=0; i<ret[0].length; i++) { ret0[i] = ret[0][i]; }
		return ret0;
	}

	/* Run multiple with the default maxN, doprint, and nruns. */
	private static double[][] Run(Belief prior, SamplingRule policy, StoppingRuleSeq stopSeq, double logc[]) {
		int nruns = 100000; // for production
		//int nruns = 1000; // for testing
		int maxN = 2000;
		return Run(prior,policy,stopSeq,logc,true,nruns,maxN);
	}

	/* Run multiple. The most general version. */
	private static double[][] Run(Belief prior, SamplingRule policy, StoppingRuleSeq stopSeq, double logc[], Boolean doprint, int nruns, int maxN) {
		double result[][];

		if (doprint) {
			/* The description of truth will have newlines in it, so we
			 * have to replace these with comments. */
			String priorDescrip = "# " + prior.toString();
			priorDescrip = priorDescrip.replaceAll("\n", "\n# ");

			System.out.println(priorDescrip);
			System.out.printf("# nruns=%d maxN=%d\n", nruns, maxN);
			System.out.println("# Policy=" + policy);
			System.out.println("# Stopping Rule=" + stopSeq);
			System.out.println("# E[N]\t\tE[OC]\t\t\ttimeouts\tc\t\tE[OC + cN]\t");
		}

		try {
			result = Simulator.LossWhenStopped(policy, prior, stopSeq, nruns, maxN);
		} catch (Exception e) {
			System.err.println("Internal Error");
			e.printStackTrace();
			result = new double[0][0];
			return result;
		}
			
		double ret[][] = new double[stopSeq.Length()][7];
		for (int s=0; s<stopSeq.Length(); s++) {
			double N = result[s][1];
			double N_stderr = result[s][4];
			double oc = result[s][0];
			double oc_stderr = result[s][3];
			double timeout = result[s][2];
			double c = StrictMath.exp(logc[s]);
			double loss = c*N+oc;
			double loss_stderr = Math.sqrt(Math.pow(c*N_stderr,2) + Math.pow(oc_stderr,2));
			if (doprint)
				System.out.printf("%.3f+/-%.3f\t%.6f+/-%.6f\t%f\t%f\t%f+/-%f\n",
						N,N_stderr,oc,oc_stderr,timeout,c,loss,loss_stderr);
			ret[s][0] = N;
			ret[s][1] = N_stderr;
			ret[s][2] = oc;
			ret[s][3] = oc_stderr;
			ret[s][4] = timeout;
			ret[s][5] = loss;
			ret[s][6] = loss_stderr;
		}
		return ret;
	}

	/*
	 * Run one stopping rule/policy, and output the times at which it
	 * stopped.  These can be turned into a histogram.
	 */
	//private static double[] StopTimeHist(Belief prior, SamplingRule policy, StoppingRule stop, int nruns, int maxN) {
//
//
	//}




	public static void main(String args[]) {
		if (args.length == 1) {
			runtest(args[0]);
		} else
			System.err.println("usage: Sim200904 test_num");
	}

	/*
	 * Run test number n. This is a function named "testn", where n
	 * is filled in with the test number.  Returns 0 if the test succeeded,
	 * -1 if the test was not found, and -2 if the test was found but
	 * failed.  Also prints an error message if it failed.  It may also
	 * fail for some unexpected reason, in which case it throws that
	 * exception.
	 */
	public static void runtest(String suffix) {
		Method method = null;
		String testname = "test" + suffix;
		try {
			method = Sim200904.class.getDeclaredMethod(testname, new Class[] {});
		} catch(NoSuchMethodException e) {
			System.err.printf("%s not found\n", testname);
		}
		try {
			method.invoke(null);
		} catch(InvocationTargetException target_e) {
			target_e.getCause().printStackTrace();
			System.err.println(testname + " failed");
		} catch(IllegalAccessException e) {
			e.printStackTrace();
			System.err.println(testname + " access error");
		}
	}


	/*
	 * Production experiments:
	 */

	/*
	 * Table 1: For several different stopping rules, $k=1$, $t_0=1$
	 * (which is small), and $\sigma=10^5$, $c=1$ (which are standard), 
	 * we want to output E[N], E[OC], E[Reward], and % suboptimality.
	 * The stopping rules to perform are:
	 *	optimal (using analytic approximation)
	 *	KG (beta=1)
	 *	\KGtilde(*) (uses analytic approximation)
	 *
	 * Other ones to possibly include:
	 *	EOC_{c,1} (what value of beta?)
	 *	EOC_{c,k} (what value of beta?)
	 *	EOCtilde_{c,k}(*)
	 * (Are these the same as KG when k=1?)
	 *
	 * Table 2: Same as Table 1, except $t_0=100$ (which is big).
	 */
	private static void testP1() throws Exception { ProductionTable(1); }
	private static void testP2() throws Exception { ProductionTable(100); }

	/*
	 * nruns=1e5 took 9.39 seconds on 5/18/2009.
	 */
	private static void ProductionTable(double t0) throws Exception {
		double logc = 0; // c=1
		int maxN = (int)1e4;
		int nruns;

		if (do_production)
		  nruns = (int)1e6; // for production.
		else
		  nruns = (int)1e5; // for quicker production.


		Belief prior = StandardK1(t0);
		SamplingRule sampler = new MaxVarianceExplore(prior); // sample only the unknown alternative.

		StoppingRule[] stop = new StoppingRule[5]; 
		String[] name = new String[5]; 
		stop[0] = new OptimalStoppingRule(prior,logc); // must be index 0.
		name[0] = "optimal";
		stop[1] = new EOCck(prior,Math.exp(logc),8);
		name[1] = "EOC_{ck}";
		stop[2] = new EOCc1(prior,Math.exp(logc),8);
		name[2] = "EOC_{c1}";
		stop[3] = new KGEnvelopeStoppingRule(prior,logc);
		name[3] = "KGE\t";
		stop[4] = new KGStoppingRule(prior,logc);
		name[4] = "KG\t";

		System.out.printf("# ProductionTable(t_0=%g)\n", t0);
		System.out.println("# k=1, c=1, \\sigma=10^5");
		System.out.printf("# nruns=%d\n", nruns);
		System.out.println("# StoppingRule\tE[N]\t\tE[OC]\tE[Reward]\t% suboptimality");

		double optimal_reward = 0; // this will be set on i=0.
		for (int i=0; i<stop.length; i++) {
			double[] ret = Run(prior, sampler, stop[i], logc, false, nruns, maxN);

			double N = ret[0];
			double N_stderr = ret[1];
			double oc = ret[2];
			double oc_stderr = ret[3];
			double timeout = ret[4];
			double reward = prior.ExpMax() - ret[5];
			double reward_stderr = ret[6];

			if (i == 0)
				optimal_reward = reward;

			double suboptimality = 100 * (optimal_reward - reward) / optimal_reward;

			System.out.printf("%s:\t%.2f+/-%.2f\t%.1f+/-%.1f\t%.1f+/-%.1f\t%.2f\n",
						name[i],
						N,N_stderr,
						oc,oc_stderr,
						reward,reward_stderr,
						suboptimality);
			if (timeout > 0) {
				System.out.printf("timeouts=%g\n", timeout);
				throw new Exception("Number of measurements exceeded maxN.  Increase maxN.");
			}

		}
	}

	/*
	 *
	 */
	private static void StoppingBoundaryForProductionTable(double t0) throws Exception {
		double logc = 0; // c=1
		int maxN = (int)1e4;
		int nruns = (int)1e6; // for production

		Belief prior = StandardK1(t0);
		SamplingRule sampler = new MaxVarianceExplore(prior); // sample only the unknown alternative.

		StoppingRule[] stop = new StoppingRule[4]; 
		String[] name = new String[4]; 
		stop[0] = new OptimalStoppingRule(prior,logc); // must be index 0.
		name[0] = "optimal";
		stop[1] = new EOCck(prior,Math.exp(logc),9);
		name[1] = "EOC_{ck}";
		stop[2] = new KGEnvelopeStoppingRule(prior,logc);
		name[2] = "KGE\t";
		stop[3] = new KGStoppingRule(prior,logc);
		name[3] = "KG\t";

		System.out.printf("# ProductionTable(t_0=%g)\n", t0);
		System.out.println("# k=1, c=1, \\sigma=10^5");
		System.out.printf("# nruns=%d\n", nruns);
		System.out.println("# StoppingRule\tE[N]\t\tE[OC]\tE[Reward]\t% suboptimality");

		double optimal_reward = 0; // this will be set on i=0.
		for (int i=0; i<stop.length; i++) {
			double[] ret = Run(prior, sampler, stop[i], logc, false, nruns, maxN);

			double N = ret[0];
			double N_stderr = ret[1];
			double oc = ret[2];
			double oc_stderr = ret[3];
			double timeout = ret[4];
			double reward = prior.ExpMax() - ret[5];
			double reward_stderr = ret[6];

			if (i == 0)
				optimal_reward = reward;

			double suboptimality = 100 * (optimal_reward - reward) / optimal_reward;

			System.out.printf("%s:\t%.2f+/-%.2f\t%.1f+/-%.1f\t%.1f+/-%.1f\t%.2f\n",
						name[i],
						N,N_stderr,
						oc,oc_stderr,
						reward,reward_stderr,
						suboptimality);
			if (timeout > 0) {
				System.out.printf("timeouts=%g\n", timeout);
				throw new Exception("Number of measurements exceeded maxN.  Increase maxN.");
			}

		}
	}
	/*
	 * Table 3: For several different stopping rules, and several different
	 * values of $k$, we output E[Reward].  (We also output E[N] and E[OC],
	 * but this is won't be included in the paper.)
	 * 
	 * The parameters used are $t_0=1$, $\sigma=10^5$ and $c=1$.  These are
	 * standard. 
	 *
	 * The stopping rules to perform are:
	 *	KG (beta=1)
	 *	\KGtilde(*) (uses analytic approximation)
	 *
	 * Other ones to possibly include:
	 *	EOC_{c,1} (what value of beta?)
	 *	EOC_{c,k} (what value of beta?)
	 *	EOCtilde_{c,k}(*)
	 *
	 * The version of this code on 5/17 with 1e6 samples took 4 hours to
	 * run.
	 * The version of this code on 5/22 16:34 with 1e3 samples took 64
	 * seconds to run, and 612 seconds for 1e4 samples.
	 *
	 * Running LL/EOC_{ck}, KGE/EOC_{ck} with method=6 for 10^4 samples
	 * took 900 seconds.
	 */
	private static void testP3b() throws Exception {
		int[] k = {1,2,3,5,10};
		doTestP3(k);
	}
	private static void testP3c() throws Exception {
		int[] k = {20};
		doTestP3(k);
	}
	private static void testP3bc() throws Exception {
		int[] k = {1,2,3,5,10,20};
		doTestP3(k);
	}


	private static void testCompareEOC() throws Exception {
		int nruns = (int)1e3; // for testing
		int maxN = (int)1e5;
		double logc = 0; // c=1
		int[] k = {5};

		int nPolicies = 2;
		BayesStoppingRuleFactory[] stoppingList = new BayesStoppingRuleFactory[nPolicies]; 
		BayesSamplingRuleFactory[] samplingList = new BayesSamplingRuleFactory[nPolicies];
		String[] nameList = new String[nPolicies];
		int policyNum = 0;
		
		/*
		stoppingList[policyNum] = new EOCckFactory(Math.exp(logc),8);
		samplingList[policyNum] = new KGESamplingRuleFactory();
		nameList[policyNum++] = "KGE/EOC_{ck}(argmax_alloc,method=8)";

		stoppingList[policyNum] = new EOCckFactory(Math.exp(logc),6);
		samplingList[policyNum] = new KGESamplingRuleFactory();
		nameList[policyNum++] = "KGE/EOC_{ck}(argmax_alloc,method=6)";


		stoppingList[policyNum] = new EOCckFactory(Math.exp(logc),8);
		samplingList[policyNum] = new LLseqArgmaxAllocFactory(Math.exp(logc),8);
		nameList[policyNum++] = "LL(method=8)/EOC_{ck}(argmax_alloc,method=8)";

		stoppingList[policyNum] = new EOCckFactory(Math.exp(logc),6);
		samplingList[policyNum] = new LLseqArgmaxAllocFactory(Math.exp(logc),6);
		nameList[policyNum++] = "LL(method=6)/EOC_{ck}(argmax_alloc,method=6)";
		*/

		stoppingList[policyNum] = new EOCckFactory(Math.exp(logc),8);
		samplingList[policyNum] = new LLseqArgmaxAllocFactory(Math.exp(logc),6);
		nameList[policyNum++] = "LL(method=6)/EOC_{ck}(argmax_alloc,method=8)";

		stoppingList[policyNum] = new EOCckFactory(Math.exp(logc),6);
		samplingList[policyNum] = new LLseqArgmaxAllocFactory(Math.exp(logc),8);
		nameList[policyNum++] = "LL(method=8)/EOC_{ck}(argmax_alloc,method=6)";




		/*
		stoppingList[policyNum] = new EOCckFactory(Math.exp(logc),5);
		samplingList[policyNum] = new LLseqArgmaxAllocFactory(Math.exp(logc),5);
		nameList[policyNum++] = "LL/EOC_{ck}(argmax_alloc,method=5)";

		stoppingList[policyNum] = new EOCckFactory(Math.exp(logc),4);
		samplingList[policyNum] = new LLseqArgmaxAllocFactory(Math.exp(logc),4);
		nameList[policyNum++] = "LL/EOC_{ck}(argmax_alloc,method=4)";

		stoppingList[policyNum] = new EOCckFactory(Math.exp(logc),3);
		samplingList[policyNum] = new LLseqArgmaxAllocFactory(Math.exp(logc),3);
		nameList[policyNum++] = "LL/EOC_{ck}(argmax_alloc,method=3)";

		stoppingList[policyNum] = new EOCckFactory(Math.exp(logc),2);
		samplingList[policyNum] = new LLseqArgmaxAllocFactory(Math.exp(logc),2);
		nameList[policyNum++] = "LL/EOC_{ck}(argmax_alloc,method=2)";

		stoppingList[policyNum] = new EOCckFactory(Math.exp(logc),1);
		samplingList[policyNum] = new LLseqArgmaxAllocFactory(Math.exp(logc),1);
		nameList[policyNum++] = "LL/EOC_{ck}(argmax_alloc,method=1)";
		*/

		/*
		stoppingList[policyNum] = new EOCc1Factory(Math.exp(logc),8);
		samplingList[policyNum] = new LLseqArgmaxAllocFactory(Math.exp(logc),8);
		nameList[policyNum++] = "LLseqc1/EOC_{c1}(argmax_alloc,method=8)";
		*/

		if (policyNum != nPolicies)
			throw new Exception("Correct policyNum to the number of policies and recompile.");

		double noise_sigma = 1e5;
		double noise_beta = 1/(noise_sigma*noise_sigma);
		double beta0 = noise_beta; // t0=1

		System.out.println("P3");
		System.out.printf("# nruns=%d maxN=%d\n", nruns, maxN);
		System.out.println("# logc=" + logc);
		System.out.println("# Policy\tk\tE[N]\t\tE[OC]\t\t\tE[Reward]\t");

		for (int i=0; i<k.length; i++) {
			for (int j=0; j<nPolicies; j++) {
				Belief prior = Belief.OneKnownAlternative(k[i],beta0,noise_beta);
				StoppingRule stop = stoppingList[j].CreateStoppingRule(prior);
				SamplingRule sample = samplingList[j].CreateSamplingRule(prior);

				//System.out.println("Allocation Rule=" + sample);
				//System.out.println("Stopping Rule=" + stop);

				double[] ret = Run(prior, sample, stop, logc, false, nruns, maxN);
				double N = ret[0];
				double N_stderr = ret[1];
				double oc = ret[2];
				double oc_stderr = ret[3];
				double timeout = ret[4];
				double reward = prior.ExpMax() - ret[5];
				double reward_stderr = ret[6];

				System.out.printf("%s\t%d\t%.2f+/-%.2f\t%.1f+/-%.1f\t%.1f+/-%.1f\n",
						nameList[j], k[i],
						N,N_stderr,
						oc,oc_stderr,
						reward,reward_stderr);
				if (timeout > 0) {
					System.out.printf("timeouts=%g\n", timeout);
					System.err.println("Number of measurements exceeded maxN.  Increase maxN.");
				}
			}
		}
	}


	private static void doTestP3(int[] k) throws Exception {
	  	int nruns;
		int maxN = (int)1e5;
		double logc = 0; // c=1

		if (do_production)
		  nruns = (int)1e6; // for production.
		else
		  nruns = (int)1e5; // for quicker production.

		int nPolicies = 1;
		BayesStoppingRuleFactory[] stoppingList = new BayesStoppingRuleFactory[nPolicies]; 
		BayesSamplingRuleFactory[] samplingList = new BayesSamplingRuleFactory[nPolicies];
		String[] nameList = new String[nPolicies];
		int policyNum = 0;

		/*
		stoppingList[policyNum] = new KGStoppingRuleFactory(logc);
		samplingList[policyNum] = new KGSamplingRuleFactory();
		nameList[policyNum++] = "KG/KG\t";

		stoppingList[policyNum] = new KGEStoppingRuleFactory(logc);
		samplingList[policyNum] = new KGESamplingRuleFactory();
		nameList[policyNum++] = "KGE/KGE(method=1)\t";

		stoppingList[policyNum] = new KGEStoppingRuleFactory(logc,2);
		samplingList[policyNum] = new KGESamplingRuleFactory();
		nameList[policyNum++] = "KGE/KGE(method=2)\t";

		stoppingList[policyNum] = new EOCc1Factory(Math.exp(logc),8);
		samplingList[policyNum] = new LLseqc1Factory(Math.exp(logc),8);
		nameList[policyNum++] = "LLseqc1(method=8)/EOC_{c1}(argmax_alloc,method=8)";

		stoppingList[policyNum] = new EOCckFactory(Math.exp(logc),8);
		samplingList[policyNum] = new MaxVarianceExploreFactory();
		nameList[policyNum++] = "MaxVariance/EOC_{ck}(method=8)";
		*/

		/*
		samplingList[policyNum] = new KGESamplingRuleFactory();
		stoppingList[policyNum] = new EOCckFactory(Math.exp(logc),8);
		nameList[policyNum++] = "KGE/EOC_{ck}(method=8)";

		stoppingList[policyNum] = new EOCckFactory(Math.exp(logc),8);
		samplingList[policyNum] = new LLseqArgmaxAllocFactory(Math.exp(logc),8);
		nameList[policyNum++] = "LL(method=8)/EOC_{ck}(argmax_alloc,method=8)";

		stoppingList[policyNum] = new DiffusionStoppingRuleFactory(Math.exp(logc));
		samplingList[policyNum] = new DiffusionSamplingRuleFactory(Math.exp(logc));
		nameList[policyNum++] = "DiffusionSampling/DiffusionStopping";
		*/

		/*
		stoppingList[policyNum] = new EOCckFactory(Math.exp(logc),8);
		samplingList[policyNum] = new KGSamplingRuleFactory();
		nameList[policyNum++] = "KG/EOC_{ck}(method=8)";
		*/

		stoppingList[policyNum] = new DiffusionStoppingRuleFactory(Math.exp(logc));
		samplingList[policyNum] = new KGSamplingRuleFactory();
		nameList[policyNum++] = "KG/DiffusionStopping";






		// Other equal alloc policies
		/*
		stoppingList[policyNum] = new KGStoppingRuleFactory(logc);
		samplingList[policyNum] = new MaxVarianceExploreFactory();
		nameList[policyNum++] = "MaxVariance/KG\t";

		stoppingList[policyNum] = new KGEStoppingRuleFactory(logc,2);
		samplingList[policyNum] = new MaxVarianceExploreFactory();
		nameList[policyNum++] = "MaxVariance/KGE(method=2)\t";

		stoppingList[policyNum] = new EOCc1Factory(Math.exp(logc),8);
		samplingList[policyNum] = new MaxVarianceExploreFactory();
		nameList[policyNum++] = "MaxVariance/EOC_{c1}(argmax_alloc,method=8)";

		stoppingList[policyNum] = new EOCckFactory(Math.exp(logc),8);
		samplingList[policyNum] = new MaxVarianceExploreFactory();
		nameList[policyNum++] = "MaxVariance/EOC_{ck}(method=8)";
		*/



		// Obsolete policies
		//stoppingList[policyNum] = new EOCckFactory(Math.exp(logc),8);
		//samplingList[policyNum] = new LLseqArgmaxAllocFactory(Math.exp(logc),8);
		//nameList[policyNum++] = "LL/EOC_{ck}(argmax_alloc)";

		//samplingList[policyNum] = new KGESamplingRuleFactory();
		//stoppingList[policyNum] = new EOCckFactory(Math.exp(logc));
		//nameList[policyNum++] = "KGE/EOC_{ck}";

		//samplingList[policyNum] = new LLseqArgmaxAllocFactory(Math.exp(logc));
		//stoppingList[policyNum] = new KGEStoppingRuleFactory(logc);
		//nameList[policyNum++] = "LL/KGE";

		//stoppingList[policyNum] = new EOCckFactory(Math.exp(logc),6);
		//samplingList[policyNum] = new LLseqArgmaxAllocFactory(Math.exp(logc),6);
		//nameList[policyNum++] = "LL(method=6)/EOC_{ck}(argmax_alloc,method=6)";

		//stoppingList[policyNum] = new EOCc1Factory(Math.exp(logc),8);
		//samplingList[policyNum] = new LLseqArgmaxAllocFactory(Math.exp(logc),8);
		//nameList[policyNum++] = "LL(method=8)/EOC_{c1}(argmax_alloc,method=8)";

		//stoppingList[policyNum] = new EOCckFactory(Math.exp(logc),8);
		//samplingList[policyNum] = new LLseqArgmaxAllocFactory(Math.exp(logc),6);
		//nameList[policyNum++] = "LL(method=6)/EOC_{ck}(argmax_alloc,method=8)";



		if (policyNum != nPolicies)
			throw new Exception("Correct policyNum to the number of policies and recompile.");




		double noise_sigma = 1e5;
		double noise_beta = 1/(noise_sigma*noise_sigma);
		double beta0 = noise_beta; // t0=1

		System.out.println("P3");
		System.out.printf("# nruns=%d maxN=%d\n", nruns, maxN);
		System.out.println("# logc=" + logc);
		System.out.println("# Policy\tk\tE[N]\t\tE[OC]\t\t\tE[Reward]\t");

		for (int i=0; i<k.length; i++) {
			for (int j=0; j<nPolicies; j++) {
				Belief prior = Belief.OneKnownAlternative(k[i],beta0,noise_beta);
				StoppingRule stop = stoppingList[j].CreateStoppingRule(prior);
				SamplingRule sample = samplingList[j].CreateSamplingRule(prior);

				//System.out.println("Allocation Rule=" + sample);
				//System.out.println("Stopping Rule=" + stop);

				double[] ret = Run(prior, sample, stop, logc, false, nruns, maxN);
				double N = ret[0];
				double N_stderr = ret[1];
				double oc = ret[2];
				double oc_stderr = ret[3];
				double timeout = ret[4];
				double reward = prior.ExpMax() - ret[5];
				double reward_stderr = ret[6];

				System.out.printf("%s\t%d\t%.2f+/-%.2f\t%.1f+/-%.1f\t%.1f+/-%.1f\n",
						nameList[j], k[i],
						N,N_stderr,
						oc,oc_stderr,
						reward,reward_stderr);
				if (timeout > 0) {
					System.out.printf("timeouts=%g\n", timeout);
					System.err.println("Number of measurements exceeded maxN.  Increase maxN.");
				}
			}
		}
	}

	/* Calculates OERmin and OERmax correspanding to testP3(). */ 
	private static void testP3a() throws Exception {
		double logc = 0; // c=1
		int[] k = {1,2,3,5,10,15,20,50,100};

		double noise_sigma = 1e5;
		double noise_beta = 1/(noise_sigma*noise_sigma);
		double beta0 = noise_beta; // t0=1

		System.out.println("P3");
		System.out.println("# logc=" + logc);
		System.out.println("# Approximation\tk\tE[N]\t\tE[N]+E[OC]\t\t\tE[Reward]\t");

		for (int i=0; i<k.length; i++) {
			Belief prior = Belief.OneKnownAlternative(k[i],beta0,noise_beta);
			LLck llc = new LLck(prior,Math.exp(logc),4); // method 4 is brute force

			double maxoer = prior.ExpMax();
			double minoer = llc.ExactNetAllocValue(llc.GoodAlloc());

			System.out.printf("maxOER\t%d\t%.0f\t%.0f\t%.0f\n",
					k[i], 0.0, 0.0, maxoer);
			System.out.printf("minOER\t%d\t%.0f\t%.0f\t%.0f\n",
					k[i], llc.GoodBudget(), maxoer-minoer, minoer);
		}
	}


	private static void testP4() throws Exception {
		int[] k = {50,100};
		doTestP4(k);
	}


	/*
	 * P4 is like P3, but for a smaller number of fast stopping rules.  It
	 * is designed for large numbers of alternatives, e.g., k=50 or 100.
	 */
	private static void doTestP4(int[] k) throws Exception {
		int nruns = (int)1e5; // for fast production
		//int nruns = (int)1e2; // for testing.
		int maxN = (int)1e7;
		double logc = 0; // c=1

		int nPolicies = 1;
		BayesStoppingRuleFactory[] stoppingList = new BayesStoppingRuleFactory[nPolicies]; 
		BayesSamplingRuleFactory[] samplingList = new BayesSamplingRuleFactory[nPolicies];
		String[] nameList = new String[nPolicies];
		int policyNum = 0;

		/*
		// Already ran these on 10.16.2009
		samplingList[policyNum] = new KGESamplingRuleFactory();
		stoppingList[policyNum] = new EOCckFactory(Math.exp(logc),8);
		nameList[policyNum++] = "KGE/EOC_{ck}(method=8)";

		stoppingList[policyNum] = new DiffusionStoppingRuleFactory(Math.exp(logc));
		samplingList[policyNum] = new DiffusionSamplingRuleFactory(Math.exp(logc));
		nameList[policyNum++] = "DiffusionSampling/DiffusionStopping";
		*/

		/* Already ran these
		stoppingList[policyNum] = new KGStoppingRuleFactory(logc);
		samplingList[policyNum] = new KGSamplingRuleFactory();
		nameList[policyNum++] = "KG/KG\t";

		stoppingList[policyNum] = new KGEStoppingRuleFactory(logc,2);
		samplingList[policyNum] = new KGESamplingRuleFactory();
		nameList[policyNum++] = "KGE/KGE(method=2)\t";

		// somewhat slow
		stoppingList[policyNum] = new EOCc1Factory(Math.exp(logc),8);
		samplingList[policyNum] = new LLseqc1Factory(Math.exp(logc),8);
		nameList[policyNum++] = "LLseqc1(method=8)/EOC_{c1}(argmax_alloc,method=8)";

		// somewhat slow
		stoppingList[policyNum] = new EOCckFactory(Math.exp(logc),8);
		samplingList[policyNum] = new LLseqArgmaxAllocFactory(Math.exp(logc),8);
		nameList[policyNum++] = "LL(method=8)/EOC_{ck}(argmax_alloc,method=8)";

		stoppingList[policyNum] = new KGEStoppingRuleFactory(logc);
		samplingList[policyNum] = new KGESamplingRuleFactory();
		nameList[policyNum++] = "KGE/KGE(method=1)\t";

		// slow
		stoppingList[policyNum] = new EOCckFactory(Math.exp(logc),8);
		samplingList[policyNum] = new MaxVarianceExploreFactory();
		nameList[policyNum++] = "MaxVariance/EOC_{ck}(method=8)";

		stoppingList[policyNum] = new DiffusionStoppingRuleFactory(Math.exp(logc));
		samplingList[policyNum] = new KGSamplingRuleFactory();
		nameList[policyNum++] = "KG/DiffusionStopping";

		*/

		stoppingList[policyNum] = new EOCckFactory(Math.exp(logc),8);
		samplingList[policyNum] = new KGSamplingRuleFactory();
		nameList[policyNum++] = "KG/EOC_{ck}(method=8)";



		if (policyNum != nPolicies)
			throw new Exception("Correct policyNum to the number of policies and recompile.");




		double noise_sigma = 1e5;
		double noise_beta = 1/(noise_sigma*noise_sigma);
		double beta0 = noise_beta; // t0=1

		System.out.println("P4");
		System.out.printf("# nruns=%d maxN=%d\n", nruns, maxN);
		System.out.println("# logc=" + logc);
		System.out.println("# Policy\tk\tE[N]\t\tE[OC]\t\t\tE[Reward]\t");

		for (int i=0; i<k.length; i++) {
			for (int j=0; j<nPolicies; j++) {
				Belief prior = Belief.OneKnownAlternative(k[i],beta0,noise_beta);
				StoppingRule stop = stoppingList[j].CreateStoppingRule(prior);
				SamplingRule sample = samplingList[j].CreateSamplingRule(prior);

				//System.out.println("Allocation Rule=" + sample);
				//System.out.println("Stopping Rule=" + stop);

				double[] ret = Run(prior, sample, stop, logc, false, nruns, maxN);
				double N = ret[0];
				double N_stderr = ret[1];
				double oc = ret[2];
				double oc_stderr = ret[3];
				double timeout = ret[4];
				double reward = prior.ExpMax() - ret[5];
				double reward_stderr = ret[6];

				System.out.printf("%s\t%d\t%.2f+/-%.2f\t%.1f+/-%.1f\t%.1f+/-%.1f\n",
						nameList[j], k[i],
						N,N_stderr,
						oc,oc_stderr,
						reward,reward_stderr);
				if (timeout > 0) {
					System.out.printf("timeouts=%g\n", timeout);
					System.err.println("Number of measurements exceeded maxN.  Increase maxN.");
				}
			}
		}
	}



	private static void testP5() throws Exception {
		double logc = 0; // c=1

	  	int nPolicies = 1;
		BayesStoppingRuleFactory[] stoppingList = new BayesStoppingRuleFactory[nPolicies]; 
		BayesSamplingRuleFactory[] samplingList = new BayesSamplingRuleFactory[nPolicies];
		String[] nameList = new String[nPolicies];
		int policyNum = 0;


	  	/*
		stoppingList[policyNum] = new KGStoppingRuleFactory(logc);
		samplingList[policyNum] = new KGSamplingRuleFactory();
		nameList[policyNum++] = "KG/KG\t";

		stoppingList[policyNum] = new KGEStoppingRuleFactory(logc);
		samplingList[policyNum] = new KGESamplingRuleFactory();
		nameList[policyNum++] = "KGE/KGE(method=1)\t";

		// slow
		stoppingList[policyNum] = new KGEStoppingRuleFactory(logc,2);
		samplingList[policyNum] = new KGESamplingRuleFactory();
		nameList[policyNum++] = "KGE/KGE(method=2)\t";

		stoppingList[policyNum] = new EOCc1Factory(Math.exp(logc),8);
		samplingList[policyNum] = new LLseqc1Factory(Math.exp(logc),8);
		nameList[policyNum++] = "LLseqc1(method=8)/EOC_{c1}(argmax_alloc,method=8)";
		*/


		/*
		// slow
		stoppingList[policyNum] = new EOCckFactory(Math.exp(logc),8);
		samplingList[policyNum] = new MaxVarianceExploreFactory();
		nameList[policyNum++] = "MaxVariance/EOC_{ck}(method=8)";
		*/

		/*
		stoppingList[policyNum] = new EOCckFactory(Math.exp(logc),8);
		samplingList[policyNum] = new KGSamplingRuleFactory();
		nameList[policyNum++] = "KG/EOC_{ck}(method=8)";
		*/

		/*
		stoppingList[policyNum] = new EOCckFactory(Math.exp(logc),8);
		samplingList[policyNum] = new KGESamplingRuleFactory();
		nameList[policyNum++] = "KGE/EOC_{ck}(method=8)";
		*/

		/*
		// slow
		stoppingList[policyNum] = new EOCckFactory(Math.exp(logc),8);
		samplingList[policyNum] = new LLseqArgmaxAllocFactory(Math.exp(logc),8);
		nameList[policyNum++] = "LL(method=8)/EOC_{ck}(argmax_alloc,method=8)";
		*/

		/*
		stoppingList[policyNum] = new EOCckFactory(Math.exp(logc),10);
		samplingList[policyNum] = new LLseqArgmaxAllocFactory(Math.exp(logc),10);
		nameList[policyNum++] = "LL(method=10)/EOC_{ck}(argmax_alloc,method=10)";
		*/

		/*
		stoppingList[policyNum] = new DiffusionStoppingRuleFactory(Math.exp(logc));
		samplingList[policyNum] = new KGSamplingRuleFactory();
		nameList[policyNum++] = "KG/DiffusionStopping";
		*/

		/*
		stoppingList[policyNum] = new DiffusionStoppingRuleFactory(Math.exp(logc));
		samplingList[policyNum] = new KGESamplingRuleFactory();
		nameList[policyNum++] = "KGE/DiffusionStopping";
		*/

		/*
		stoppingList[policyNum] = new DiffusionStoppingRule2Factory(Math.exp(logc));
		samplingList[policyNum] = new KGESamplingRuleFactory();
		nameList[policyNum++] = "KGE/DiffusionStopping2";
		*/
	
		/*
		// slow
		stoppingList[policyNum] = new DiffusionStoppingRuleFactory(Math.exp(logc));
		samplingList[policyNum] = new LLseqArgmaxAllocFactory(Math.exp(logc),8);
		nameList[policyNum++] = "LL(method=8)/DiffusionStopping";
		*/

		/*
		stoppingList[policyNum] = new DiffusionStoppingRuleFactory(Math.exp(logc));
		samplingList[policyNum] = new DiffusionSamplingRuleFactory(Math.exp(logc));
		nameList[policyNum++] = "DiffusionSampling/DiffusionStopping";
		*/


		/* These are the extra method=10 things. */
		/*
		stoppingList[policyNum] = new EOCc1Factory(Math.exp(logc),10);
		samplingList[policyNum] = new LLseqc1Factory(Math.exp(logc),10);
		nameList[policyNum++] = "LLseqc1(method=10)/EOC_{c1}(argmax_alloc,method=10)";
		*/

		stoppingList[policyNum] = new EOCckFactory(Math.exp(logc),10);
		samplingList[policyNum] = new MaxVarianceExploreFactory();
		nameList[policyNum++] = "MaxVariance/EOC_{ck}(method=10)";

		/* 
		stoppingList[policyNum] = new EOCckFactory(Math.exp(logc),10);
		samplingList[policyNum] = new KGSamplingRuleFactory();
		nameList[policyNum++] = "KG/EOC_{ck}(method=10)";
		*/

		/* IN PROGRESS
		stoppingList[policyNum] = new EOCckFactory(Math.exp(logc),10);
		samplingList[policyNum] = new KGESamplingRuleFactory();
		nameList[policyNum++] = "KGE/EOC_{ck}(method=10)";
		*/




		if (policyNum != nPolicies)
			throw new Exception("Correct policyNum to the number of policies and recompile.");

		int[] smallk = {2,3,5,10,20};
		int smallk_nruns = (int)1e3;
		//int smallk_nruns = (int)1e6;
		doTestP5(smallk,smallk_nruns,stoppingList,samplingList,nameList);

		int[] bigk = {50,100};
		int bigk_nruns = (int)1e2;
		//int bigk_nruns = (int)1e5;
		doTestP5(bigk,bigk_nruns,stoppingList,samplingList,nameList);
	}


	private static void doTestP5(int[] k, int nruns, BayesStoppingRuleFactory[] stoppingList, BayesSamplingRuleFactory[] samplingList, String[] nameList) throws Exception {
		double logc = 0; // c=1
		int maxN = (int)1e7;

		int nPolicies = stoppingList.length;
		if (nPolicies != samplingList.length || nPolicies != nameList.length)
		  throw new Exception("Inputs to doTestP5 (stoppingList, samplingList, nameList) have different lengths."); 

		double noise_sigma = 1e5;
		double noise_beta = 1/(noise_sigma*noise_sigma);
		double beta0 = noise_beta; // t0=1

		System.out.println("P5");
		System.out.printf("# nruns=%d maxN=%d\n", nruns, maxN);
		System.out.println("# logc=" + logc);
		System.out.println("# Policy\tk\tE[N]\t\tE[OC]\t\t\tE[Reward]\t");

		for (int i=0; i<k.length; i++) {
			for (int j=0; j<nPolicies; j++) {
				Belief prior = Belief.OneKnownAlternative(k[i],beta0,noise_beta);
				StoppingRule stop = stoppingList[j].CreateStoppingRule(prior);
				SamplingRule sample = samplingList[j].CreateSamplingRule(prior);

				//System.out.println("Allocation Rule=" + sample);
				//System.out.println("Stopping Rule=" + stop);

				double[] ret = Run(prior, sample, stop, logc, false, nruns, maxN);
				double N = ret[0];
				double N_stderr = ret[1];
				double oc = ret[2];
				double oc_stderr = ret[3];
				double timeout = ret[4];
				double reward = prior.ExpMax() - ret[5];
				double reward_stderr = ret[6];

				System.out.printf("%s\t%d\t%.2f+/-%.2f\t%.1f+/-%.1f\t%.1f+/-%.1f\n",
						nameList[j], k[i],
						N,N_stderr,
						oc,oc_stderr,
						reward,reward_stderr);
				if (timeout > 0) {
					System.out.printf("timeouts=%g\n", timeout);
					System.err.println("Number of measurements exceeded maxN.  Increase maxN.");
				}
			}
		}
	}


	/*
	 * This does a fixed horizon version of P3.
	 */
	private static void test6() throws Exception {
		int nruns = (int)1e5; // for production.
		//int nruns = 1000; // for testing
		int maxN = (int)1e4;
		double logc = 0; // c=1
		int[] k = {1,2,3,4,5,6,7,8,9,10,12,15,20};
		//int[] k = {1};
		int nPolicies = 3;
		BayesSamplingRuleFactory[] samplingList = new BayesSamplingRuleFactory[nPolicies];
		String[] nameList = new String[nPolicies];
		int horizon = 30;

		samplingList[0] = new KGSamplingRuleFactory();
		nameList[0] = "KG/KG\t";
		samplingList[1] = new KGESamplingRuleFactory();
		nameList[1] = "KGE/KGE\t";
		samplingList[2] = new LLseqArgmaxAllocFactory(Math.exp(logc));
		nameList[2] = "LL/EOC_{ck}";

		double noise_sigma = 1e5;
		double noise_beta = 1/(noise_sigma*noise_sigma);
		double beta0 = noise_beta; // t0=1

		System.out.println("test6");
		System.out.printf("# nruns=%d N=%d\n", nruns, horizon);
		System.out.println("# logc=" + logc);
		System.out.println("# Policy\tk\tE[N]\t\tE[OC]\t\t\tE[Reward]\t");

		for (int i=0; i<k.length; i++) {
			for (int j=0; j<nPolicies; j++) {
				Belief prior = Belief.OneKnownAlternative(k[i],beta0,noise_beta);
				StoppingRule stop = new FixedStop(horizon);
				SamplingRule sample = samplingList[j].CreateSamplingRule(prior);

				//System.out.println("Allocation Rule=" + sample);
				//System.out.println("Stopping Rule=" + stop);

				double[] ret = Run(prior, sample, stop, logc, false, nruns, maxN);
				double N = ret[0];
				double N_stderr = ret[1];
				double oc = ret[2];
				double oc_stderr = ret[3];
				double timeout = ret[4];
				double reward = prior.ExpMax() - ret[5];
				double reward_stderr = ret[6];

				System.out.printf("%s\t%d\t%.2f+/-%.2f\t%.1f+/-%.1f\t%.1f+/-%.1f\n",
						nameList[j], k[i],
						N,N_stderr,
						oc,oc_stderr,
						reward,reward_stderr);
				if (timeout > 0) {
					System.out.printf("timeouts=%g\n", timeout);
					throw new Exception("Number of measurements exceeded maxN.  Increase maxN.");
				}
			}
		}
	}

	/*
	 * Run policies like KG for multiple values of the cost.  Report E[N]
	 * E[OC], and E[Reward] (for c=1).
	 */
	private static void testP5a() throws Exception { RunVaryC("KG(1)", 100, 1); }
	private static void testP5b() throws Exception { RunVaryC("KG(*)", 100, 1); }
	private static void testP5c() throws Exception { RunVaryC("optimal", 100, 1); }
	private static void testP5d() throws Exception { RunVaryC("EOC(ck)", 100, 1); }

	private static void testP6a() throws Exception { RunVaryC("KG(1)", 1, 3); }
	private static void testP6b() throws Exception { RunVaryC("KG(*)", 1, 3); }
	private static void testP6d() throws Exception { RunVaryC("EOC(ck)", 1, 3); }

	private static void RunVaryC(String policyName, double t0, int k) throws Exception {
	  double real_c = 1;
	  int maxN = 100000;
	  int nruns = 1000000; // production
	  // int nruns = 1000;
	  int n_logc;
	  double logc[];

	  Belief prior = StandardK(t0, k);
	  SamplingRule sampler;
	  if (k == 1)
	    sampler = new MaxVarianceExplore(prior);
	  else
	    sampler = new KG(prior);

	  StoppingRuleSeq stopSeq;
	  int n_points = 100;
	  logc = new double[n_points];
	  double min_logc;
	  if (policyName.equals("KG(1)")) {
	    min_logc = -500;
	    for (int n=0; n<n_points; logc[n++] = n*min_logc/n_points);
	    stopSeq = KGStoppingRule.Seq(prior, logc);
	  } else if (policyName.equals("KG(*)")) {
	    min_logc = -5;
	    for (int n=0; n<n_points; logc[n++] = n*min_logc/n_points);
	    stopSeq = KGEnvelopeStoppingRule.Seq(prior, logc);
	  } else if (policyName.equals("optimal")) {
	    min_logc = -5;
	    for (int n=0; n<n_points; logc[n++] = n*min_logc/n_points);
	    stopSeq = new OptimalStoppingRuleSeq(prior, logc);
	    /*
	  } else if (policyName.equals("EOC(ck)")) {
	    min_logc = -5;
	    for (int n=0; n<n_points; logc[n++] = n*min_logc/n_points);
	    stopSeq = EOCckStoppingRule.Seq(prior, logc);
	  } else if (policyName.equals("EOC(c1)")) {
	    min_logc = -5;
	    for (int n=0; n<n_points; logc[n++] = n*min_logc/n_points);
	    stopSeq = EOCc1StoppingRule.Seq(prior, logc);
	    */
	  } else {
	    throw new Exception("Policy name unknown: " + policyName);
	  }

	// logc[n] = max_logc - n*(max_logc-min_logc)/n_points


	  /*
	     stop[0] = new OptimalStoppingRule(prior,logc); // must be index 0.
	     name[0] = "optimal";
	     */

	  System.out.printf("# %s nruns=%d\n", policyName, nruns);
	  System.out.println("# log(c)\tOC\t\t\tN\t\tReward(c=1)");
	  double ret[][];
	  ret = Simulator.LossWhenStopped(sampler, prior, stopSeq, nruns, maxN);
	  for (int s=0; s<stopSeq.Length(); s++) {
	      double oc = ret[s][0];
	      double N = ret[s][1];
	      double oc_std = ret[s][3];
	      double N_std = ret[s][4];
	      int timeout = (int)ret[s][2];
	      double loss = real_c*N+oc;
	      double loss_std = Math.sqrt(Math.pow(real_c*N_std,2) + Math.pow(oc_std,2));
	      double reward = prior.ExpMax() - loss;
	      double reward_std = loss_std;

	      System.out.printf("%f\t%.2f +/- %.2f\t%.2f +/- %.2f\t%.2f +/- %.2f\n",
		  logc[s],
		  oc, oc_std,
		  N, N_std,
		  reward, reward_std);

	      if (timeout > 0) {
		System.out.printf("timeouts=%d\n", timeout);
		throw new Exception("Number of measurements exceeded maxN.  Increase maxN.");
	      }
	    }
	}

	/* Returns the standard k=1 belief. */
	private static Belief StandardK1(double t0) {
	  	return StandardK(t0, 1);
	}

	private static Belief StandardK(double t0, int k) {
		double noise_sigma = 1e5;
		double noise_beta = 1/(noise_sigma*noise_sigma);
		double beta0 = t0*noise_beta;
		return Belief.OneKnownAlternative(k,beta0,noise_beta);
	}


}
